/*
 * MapLine.java
 *
 * Created on September 15, 2006, 5:04 PM
 *
 */

package ch.ethz.karto.gui;

import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

/**
 * MapLine stores a series of 2D points in a vector that form a simple line.
 * @author Bernhard Jenny, Institute of Cartography, ETH Zurich.
 */
public class MapLine {

    /**
     * A vector that holds all points of this line.
     */
    private ArrayList points = new ArrayList();
    
    /**
     * The bounding box of this line.
     */
    private Rectangle2D extension = null;
    
    /**
     * The path that can be used to draw this line.
     */
    private GeneralPath path = null;

    /** Creates a new instance of MapLine */
    public MapLine() {
    }
    
    /**
     * Return the number of points of this line.
     */
    public int size() {
        return points.size();
    }
    
    /**
     * Return a copy of a point.
     * @param pointID The index of the point to return (zero-based).
     * @return A reference to the point at position pointID.
     */
    public Point2D.Double getPoint (int pointID) {
        return (Point2D.Double)points.get(pointID);
    }
    
    /**
     * Add a point at the end of the line.
     * @mapPoint The point to add.
     */
    public void addPoint(Point2D mapPoint) {
        if (Double.isNaN(mapPoint.getX()) || Double.isNaN(mapPoint.getY())) {
            return;
        }
        points.add(mapPoint);
        pointChanged();
    }
    
    /**
     * Add a point at the end of the line.
     * @param x The horizontal coordinate of the point.
     * @param y The vertical coordinate of the point.
     */
    public void addPoint(double x, double y) {
        addPoint(new Point2D.Double(x, y));
    }
    
    /**
     * Remove a point from the line.
     * @param mapPoint The point to remove.
     */
    public void removePoint(Point2D mapPoint) {
        points.remove(mapPoint);
        pointChanged();
    }
    
    /**
     * Remove a point from the line.
     * @param pointID The index of the point to remove (zero-based).
     */
    public void removePoint(int pointID) {
        points.remove(pointID);
        pointChanged();
    }
    
    /**
     * Return the bounding box of this line.
     * @return The bounding box.
     */
    public java.awt.geom.Rectangle2D getExtension() {
        
        if (points.size() == 0) {
            return null;
        }
        
        // If a bounding box has been computed before, return it.
        // This is probably faster than looping throuh all points.
        if (extension != null) {
            return extension;
        }
        
        // search the smallest and largest coordinates in x and y direction.
        double xMin = Double.MAX_VALUE;
        double xMax = -Double.MAX_VALUE;
        double yMin = Double.MAX_VALUE;
        double yMax = -Double.MAX_VALUE;
        
        int nbrPoints = points.size();
        for (int i = 0; i < nbrPoints; i++) {
            Point2D point = (Point2D)points.get(i);
            double x = point.getX();
            double y = point.getY();
            if (x < xMin)
                xMin = x;
            if (x > xMax)
                xMax = x;
            if (y < yMin)
                yMin = y;
            if (y > yMax)
                yMax = y;
        }
        
        // store the found bounding values
        extension = new java.awt.geom.Rectangle2D.Double(xMin, yMin,
                xMax - xMin, yMax - yMin);
        return extension;

    }
    
    /**
     * Return a GeneralPath that can be drawn.
     * @return The GeneralPath.
     */
    public GeneralPath getPath() {
        
        // if the path has been constructed before, return it.
        if (path != null) {
            return path;
        }
        
        path = new GeneralPath();
        int nbrPoints = size();
        
        // a line must have at least 2 points
        if (nbrPoints < 2) {
            return path;
        }
        
        // add first point
        Point2D point = (Point2D)points.get(0);
        path.moveTo((float)point.getX(), (float)point.getY());
        
        // add all following points
        for (int pointID = 1; pointID < nbrPoints; pointID++) {
            point = (Point2D)points.get(pointID);
            path.lineTo((float)point.getX(), (float)point.getY());
        }
        
        return path;
    }
    
    /**
     * Private helper method that must be called whenever a point changes.
     */
    private void pointChanged() {
        path = null;
        extension = null;
    }


}